# python3-opsi
# This code is part of the opsi.org project
# http://opsi.org
#
# Copyright (c) uib gmbh (www.uib.de)
# This sourcecode is owned by the uib gmbh, D55118 Mainz, Germany
# and published under the Terms of the GNU Affero General Public License.
# Text of the AGPL: http://www.gnu.org/licenses/agpl-3.0-standalone.html
# credits: http://www.opsi.org/credits/

# Debian
# https://www.debian.org/releases/stable/amd64/ch05s03.html.de
# https://wiki.debian.org/DebianInstaller/
# https://www.debian.org/releases/stable/amd64/apb.html.de
# preseed files:
# http://www.debian-administration.org/article/394/Automating_new_Debian_installations_with_preseeding

# Ubuntu:
#http://archive.ubuntu.com/ubuntu/dists/trusty-updates/main/installer-amd64/current/images/netboot/ubuntu-installer/amd64/
#https://wiki.ubuntu.com/UbiquityAutomation
#https://help.ubuntu.com/14.04/installation-guide/example-preseed.txt

try:
    import ConfigParser
except:
    import configparser
from OPSI.Util.File import IniFile
from OPSI.Util.File import TextFile
from OPSI.Backend.BackendManager import BackendManager
from OPSI.Backend.JSONRPC import JSONRPCBackend
from OPSI.Util import md5sum
from OPSI.Util import blowfishDecrypt
from hashlib import md5
try:
    from twisted.conch.ssh import keys
except:
    pass
try:
    from urlparse import urlparse
except:
    from urllib.parse import urlparse
from os.path import splitext, basename
import platform
import re, xml.dom.minidom, os, gzip
import crypt
import shutil
from base64 import b64encode
from os import urandom


isOpsi43 = False
if backend.backend_info()['opsiVersion'].startswith('4.3'):
    isOpsi43 = True

if isOpsi43:
    show_message(u"Installing Product: %s" % (productId), 10)

opsiVersion = backend.backend_info()['opsiVersion']
source = SCRIPT_PATH

# Do hardware inventory
logger.notice(u"Fetching opsi hw audit configuration")
hwconfig = backend.auditHardware_getConfig()
if isOpsi43:
    show_message(u"Running hardware inventory", 10)
logger.notice(u"Running hardware inventory")
auditHardwareOnHosts = auditHardware(config = hwconfig, hostId = clientId)

logger.notice(u"Sending hardware information to service")
backend.auditHardwareOnHost_setObsolete(clientId)
backend.auditHardwareOnHost_createObjects(auditHardwareOnHosts)

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsisetuplib.py','/tmp')
exec(open('/tmp/opsisetuplib.py').read())

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsipreparelib.py','/tmp')
exec(open('/tmp/opsipreparelib.py').read())

initOpsiSetupLib()

useGpt = False
inUefi = inUefiMode()
if inUefi:
    try:
        inSecureboot = isSecureboot()
    except NameError:
        inSecureboot = False

    if inSecureboot:
        logger.notice(u"we are running in Secure Boot mode")
        scriptMessageSubject.setMessage(u"we are running in Secure Boot mode")
    else:
        logger.notice(u"we are not running in Secure Boot mode")
        scriptMessageSubject.setMessage(u"we are not running in Secure Boot mode")
    useGpt = True

if useGpt:
    logger.notice(u"Will use GPT")
else:
    logger.notice(u"Will use MBR Partitions")


proxy = productProperties.get('proxy','')

nonFreeFirmware = False
if (str(productProperties.get('use_non_free_firmware_in_installer', 'true')).lower() == 'true'):
    nonFreeFirmware = True

install_opsi_client_agent = False
do_opsi_client = "false"
if (str(productProperties.get('install_opsi-client-agent', 'true')).lower() == 'true'):
    install_opsi_client_agent = True
    do_opsi_client = "true"

install_unattended = True
if (str(productProperties.get('install_unattended', 'true')).lower() == 'false'):
    install_unattended = False

installation_method = productProperties.get('installation_method', 'kexec')

locale = productProperties.get('locale','de_DE')
language = locale.split('_')[0]
country = locale.split('_')[1]
keymap = productProperties.get('console_keymap', 'de-latin1-nodeadkeys')
try:
    xkbvariant = keymap.split('-')[2]
except:
    xkbvariant = ''
    pass
xkblayout = keymap.split('-')[0]

online_repository = productProperties.get('online_repository', 'http://ftp.de.debian.org/debian')

myarchitecture = getArchitecture()
if myarchitecture == 'amd64' and 'opsi' in online_repository:
    online_repository = online_repository + '/amd64'
if myarchitecture == 'i386' and 'opsi' in online_repository:
    online_repository = online_repository + '/i386'

logger.notice(u"splitting online_repository ...")
online_repository_protocol=online_repository.split('://')[0]
online_repo_adr = online_repository.split('://')[1]
online_repository_base=online_repo_adr.split('/')[0]
# get the rest: split has num argument
online_repository_dir='/'+online_repo_adr.split('/',1)[1]

partition_disk = productProperties.get('partition_disk', 'first')
partition_method = productProperties.get('partition_method', 'lvm')
#if inUefi:
#       partition_method = 'efi'
partition_recipe = productProperties.get('partition_recipe', 'atomic')
if inUefi:
    partition_recipe = 'amd64/efi'
desktop_package = productProperties.get('desktop_package', 'standard')
#language_packs = productProperties.get('language_packs', 'standard')
upgrade_select = productProperties.get('upgrade_select', 'none')


my_broadcast = getBroadcastAddress(ipaddress, netmask)
my_network = getNetworkAddress(ipaddress, netmask)

# Get and encrypt passwords
# user
username = productProperties.get('user_name', 'user')
userpassclear = productProperties.get('user_password', 'linux123')
userpassenc = shadow_password_hash(userpassclear)

# root
rootpassclear = productProperties.get('root_password', 'linux123')
rootpassenc = shadow_password_hash(rootpassclear)

if (proxy == ''):
    do_proxy = 'false'
    proxy_addr = ''
else:
    do_proxy = 'true'
    proxy_addr = proxy
    if proxy_addr.startswith('https:'):
        execute('export HTTPS_PROXY="%s"' % proxy_addr)
    else:
        execute('export HTTP_PROXY="%s"' % proxy_addr)

opsi_client_agent_install_script = ''

# get machine architecture
myarchitecture = getArchitecture()
if myarchitecture == 'amd64':
    scriptMessageSubject.setMessage(u"Installing 64 Bit system....")
    logger.notice(u"Installing 64 Bit system....")
    myarchitectureBits = '64'
else:
    scriptMessageSubject.setMessage("Installing 32 Bit system....")
    logger.notice(u"Installing 32 Bit system....")
    myarchitectureBits = '32'

if isOpsi43:
    show_message("Preparing disk", 20)
scriptMessageSubject.setMessage("Create temporary place on disk.... ")
logger.notice(u"Create temporary place on disk.... ")
#prepareDisk()
target = '/mnt/hd'
systemPartitionNumber = 1
if inUefi:
    systemPartitionNumber = 3
    if installation_method == 'reboot':
        grubPartitionNumber = 3
        systemPartitionNumber += 1
        grubPartitionSizeM = 2000
        grubPartitionLabel = 'INSTALLER'
systemPartitionSizeM = 2000
systemPartitionLabel  = 'LINUX'
blockAlignment = True

logger.notice(u"Swapoff all to avoid disk in use .....")
execute("swapoff -a -v")
lvmClearVolumeGroups()
execute('sfdisk -l')

# Get harddisks
disks = getHarddisks()
disks = sorted(disks, key=lambda disk: disk.device)
# use first disk
disk = disks[0]
if partition_disk == "first":
    partition_disk = disk.device
else:
    disk.device = partition_disk

if useGpt:
    logger.notice(u"Clear partition table of %s" % (disk.device))
    execute('/usr/sbin/sgdisk --zap-all %s' % (disk.device))
    execute('/usr/sbin/sgdisk --clear -g -p %s' % (disk.device))
    if inUefi:
        currentbootentry = uefiGetActBootEntry()
        if currentbootentry:
            uefiNetbootLabel = currentbootentry['label']
            logger.notice('Current (net) boot entry: %s' % uefiNetbootLabel)
            backend.config_createUnicode("clientConfig.uefiNetbootLabel","clientspecific uefi firmware netboot entry label")
            backend.configState_create("clientConfig.uefiNetbootLabel", clientId, [ uefiNetbootLabel ] )
        else:
            logger.notice('No usable current (net) boot entry found')

        deleteAllEntriesFromEFIBootmgr('opsitempwinpe')
        deleteAllEntriesFromEFIBootmgr('Windows Boot Manager')
        deleteAllEntriesFromEFIBootmgr('ubuntu')
        deleteAllEntriesFromEFIBootmgr('debian')
        deleteAllEntriesFromEFIBootmgr('installer')
        deleteAllEntriesFromEFIBootmgr('opsibootimage')
        logger.notice(u"Create EFI system partition on %s1" % (disk.device))
        execute('/usr/sbin/sgdisk -g --new 1::100M -t 1:ef00 -c 1:\"EFI system partition\" --partition-guid=C12A7328-F81F-11D2-BA4B-00A0C93EC93B -p %s' % (disk.device))
        execute('/bin/sleep 2')
        execute('/usr/sbin/sgdisk -p %s' % (disk.device))
        logger.notice(u"Make file system on EFI system partition on %s1" % (disk.device))
        normalDiskDevices = ['/dev/sda', '/dev/vda', '/dev/sdb']
        if any(normalDisk in disk.device for normalDisk in normalDiskDevices):
            execute('/sbin/mkfs.vfat %s1' % (disk.device))
        else:
            execute('/sbin/mkfs.vfat %sp1' % (disk.device))
        logger.notice(u"Create MS reserved partition on %s2" % (disk.device))
        execute('/usr/sbin/sgdisk -g --new 2::+128M -t 2:0c01 -c 2:\"MS reserved\" --partition-guid=E3C9E316-0B5C-4DB8-817D-F92DF00215AE -p %s' % (disk.device))
else:
    disk.deletePartitionTable()

if installation_method == 'reboot':
    if inUefi:
        grubTarget = '/mnt/grub/'
        checkCreateDir(grubTarget)
        execute('/usr/sbin/sgdisk --info=%d %s' % (grubPartitionNumber, disk.device))
        createPartitionEx("installer", grubPartitionNumber, grubPartitionLabel, 0, grubPartitionSizeM, gptAutoAppend = True, boot = True, gptAttribute = "63" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "fat32", gptFileSystem="0700")
        execute('/usr/sbin/sgdisk -t %d:0700 --attributes %d:set:63 --attributes %d:clear:0 -p %s' % (grubPartitionNumber, grubPartitionNumber, grubPartitionNumber, disk.device))
        testPartitionEx(disk.device, grubPartitionNumber, useGpt)
        checkClearPartition("none", disk.device, grubPartitionNumber)
        createFilesystemEx("bootimage", grubPartitionNumber, "bootimage", "fat32", disk.device , useGpt)
        checkMount(grubTarget, disk.device, grubPartitionNumber, "vfat")
        if not depot.exists("%s/installfiles64/linux" % (source)):
            raise Exception("kernel for locally installed installer not found")
        logger.notice("Copying files to partition %d" % grubPartitionNumber)
        depot.copy('/%s/installfiles64/*' % (productId), grubTarget)
        localizationLine = 'locale=%s keymap=%s console-keymaps-at/keymap=%s' % (locale,keymap,keymap)
        newLines = []
        with open('%s/grub/grub.cfg' % (grubTarget), 'r+') as grubConfig:
            for line in grubConfig:
                if line.strip().startswith('linux'):
                    newLines.append('%s %s\n' % (line.rstrip(), localizationLine.rstrip()))
                    continue
                newLines.append(line)

        with open('%s/grub/grub.cfg' % (grubTarget), 'w') as grubConfig:
            grubConfig.writelines(newLines)

        execute('/bin/efibootmgr -c --disk %s --part %s -L installer --loader \\\\shimx64.efi.signed' % (disk.device, grubPartitionNumber))
        setFirstUefiBoot('installer')
        setNextUefiBoot('installer')
        execute('/bin/efibootmgr -v')
    else:
        logger.warning(u"reboot installation method only works using UEFI. Reverting back to kexec method")
        installation_method = "kexec"

fdiskexitcode = os.system('%s %s >> /dev/null 2>&1' % (which('partprobe'), disk.device))
createPartitionEx("system", systemPartitionNumber, systemPartitionLabel, 0, systemPartitionSizeM, gptAutoAppend = True, boot = False, gptAttribute = "" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "ext4", gptFileSystem="8300")
createFilesystemEx("system", systemPartitionNumber, systemPartitionLabel, "ext4", disk.device, useGpt)
systemPartitionDevicePath = getPartitionDevicePath(disk.device, systemPartitionNumber)

# Partition table deleted => delete all netboot product states
deletePocs = []
for poc in backend.productOnClient_getObjects(clientId = clientId):
    if ((poc.productType == 'NetbootProduct') and (poc.productId != productId)) or \
       ((poc.productType == 'LocalbootProduct') and (poc.installationStatus == 'not_installed') and (poc.actionRequest == 'none')):
        deletePocs.append(poc)
backend.productOnClient_deleteObjects(deletePocs)

if isOpsi43:
    show_message("Mounting partitions and copying data", 40)
# Mount temporary partition
checkCreateDir(target)
checkClearMountpoint(target, disk.device, systemPartitionNumber, systemPartitionDevicePath)
checkClearPartition(target, disk.device, systemPartitionNumber, systemPartitionDevicePath)
checkMount(target, disk.device, systemPartitionNumber, "ext4", systemPartitionDevicePath)
checkCreateDir(target+'/tmp')

scriptMessageSubject.setMessage("Getting kernel and initrd.... ")
logger.notice(u"Getting kernel and initrd.... ")

depot.copy(SCRIPT_PATH + '/installfiles%s/linux' % (myarchitectureBits), target+'/tmp')
depot.copy(SCRIPT_PATH + '/installfiles%s/initrd.gz' % myarchitectureBits, target+'/tmp')

scriptMessageSubject.setMessage("Uncompressing initrd.... ")
logger.notice(u"Uncompressing initrd.... ")
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass
execute("gunzip %s/tmp/initrd.gz"% (target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass

scriptMessageSubject.setMessage("Patching initrd.... ")
logger.notice(u"Mounting initrd ....")
execute("mkdir -p %s/tmp/mnt" % (target))
execute("cd %s/tmp/mnt ; cpio -idm -H newc < ../initrd" % (target))

logger.notice(u"Copy profile for patch....")
preseed = productProperties.get('preseed', 'auto')
if preseed == 'auto':
    logger.notice(u"Default preseed is used")
    depot.copy(SCRIPT_PATH+'/debian.cfg', target+'/tmp/installation_profile')
else:
    logger.notice(u"Custom installation profile (%s) found and will be used" % preseed)
    depot.copy(SCRIPT_PATH+'/custom/%s' % preseed, target+'/tmp/installation_profile')

logger.notice(u"Copy /etc/default/keyboard for patch....")
depot.copy(SCRIPT_PATH+'/keyboard', target+'/tmp/keyboard')


localDebianRepo = False
if 'debian.org/debian' not in online_repository:
    localDebianRepo = True

logger.notice(u"Prepare patch file....")
opsipatchscript='/sbin/opsipatchscript.sh'
with open('%s/tmp/mnt/%s' % (target,opsipatchscript), "a") as myfile:
    myfile.write('#! /bin/sh\n')
    myfile.write('set -x\n')
    myfile.write('chroot /target install-keymap %s\n' % keymap)
    myfile.write('cp /opsitmp/keyboard /target/etc/default/keyboard\n')
    #myfile.write('chroot /target cp -a /etc/default/keyboard /tmp/\n')
    #myfile.write('chroot /target cat /tmp/keyboard | sed s"/XKBLAYOUT=\".*/XKBLAYOUT=\"%s\"/" > /etc/default/keyboard\n' % language)
try:
    sourcesList = requests.get('%s/sources.list' % online_repository)
    if sourcesList.status_code == 200:
        logger.notice(u"sources.list found")
        if localDebianRepo:
            with open('%s/tmp/mnt/%s' % (target,opsipatchscript), "a") as myfile:
                myfile.write('wget -O /target/etc/apt/sources.list.d/original.list %s/sources.list\n' % online_repository)
    with open('%s/tmp/mnt/%s' % (target,opsipatchscript), "a") as myfile:
        myfile.write('in-target export DEBIAN_FRONTEND=noninteractive\n')
        myfile.write('in-target apt-get update\n')
        myfile.write('in-target apt-get install librsync2 python3 pkg-config openssl dnsutils -q --yes\n')
        myfile.write('in-target apt-get install %s -q --yes\n' % desktop_package)
except requests.exceptions.RequestException as e:
    pass

execute('chmod u+x %s/tmp/mnt/%s\n' % (target,opsipatchscript))
logger.notice('Here comes the %s/tmp/mnt/%s' % (target,opsipatchscript))
execute('cat %s/tmp/mnt/%s' % (target,opsipatchscript))

logger.notice(u"Copy SYSCONF_INI....")
# Copy sysconf.ini to opsi-client-agent config dir
copy(SYSCONF_INI , target +'/tmp/')

logger.notice(u"Prepare patch input....")
with open(PATCHA_IN, "a") as myfile:
    myfile.write('networkaddress=%s\n' % my_network)
    myfile.write('broadcastaddress=%s\n' % my_broadcast)
    myfile.write('nameserveraddress=%s\n' % nameserver)
    #myfile.write('myuserpassword=%s\n' % myuserpassword)
    #myfile.write('myrootpassword=%s\n' % myrootpassword)
    myfile.write('do_opsi_client=%s\n' % do_opsi_client)
    myfile.write('username=%s\n' % username)
    myfile.write('userpassenc=%s\n' % userpassenc)
    myfile.write('rootpassenc=%s\n' % rootpassenc)
    myfile.write('do_proxy=%s\n' % do_proxy)
    myfile.write('proxy_addr=%s\n' % proxy_addr)
    myfile.write('online_repository=%s\n' % online_repository)
    myfile.write('online_repository_protocol=%s\n' % online_repository_protocol)
    myfile.write('online_repository_base=%s\n' % online_repository_base)
    myfile.write('online_repository_dir=%s\n' % online_repository_dir)
    myfile.write('opsi_client_agent_install_script=%s\n' % opsi_client_agent_install_script)
    myfile.write('partition_disk=%s\n' % partition_disk)
    myfile.write('partition_method=%s\n' % partition_method)
    myfile.write('partition_recipe=%s\n' % partition_recipe)
    myfile.write('desktop_package=%s\n' % desktop_package)
    myfile.write('language=%s\n' % language)
    myfile.write('country=%s\n' % country)
    myfile.write('opsipatchscript=%s\n' % opsipatchscript)
    myfile.write('xkbvariant=%s\n' % xkbvariant)
    myfile.write('xkblayout=%s\n' % xkblayout)
    myfile.write('upgrade_select=%s\n' % (upgrade_select))
    myfile.write('productid=%s\n' % (productId))
    myfile.write('service_ip_address=%s\n' % (nextserver.replace("/rpc", "").rstrip("/")))
    myfile.write('depotAddress=%s\n' % (depotUrl.split('/')[2].replace(':4447','')))

logger.notice(u"Here comes the PATCHA_IN (%s):" % PATCHA_IN)
execute('cat %s' % (PATCHA_IN))

if isOpsi43:
    show_message("patching files", 90)

if not username:
    execute('sed -i "s/\#d-i passwd\/make-user/d-i passwd\/make-user/g" %s/tmp/installation_profile' % target)

logger.notice(u"Patching installation_profile")
if install_opsi_client_agent == False:
    lines = []
    with open(target + '/tmp/installation_profile', 'r') as inputpreseed:
        lines = inputpreseed.readlines()
    outlines = []
    keep = True
    for line in lines:
        if "in-target mkdir -p /var/tmp/opsi-linux-client-agent" in line:
            keep = False
        if keep:
            outlines.append(line)
        if "systemctl enable" in line:
            keep = True
    with open(target + '/tmp/installation_profile', 'w') as outputpreseed:
        outputpreseed.writelines(outlines)

for patchfile in ( target + '/tmp/installation_profile' , target+'/tmp/keyboard'):
    execute('/usr/local/bin/patcha %s' % patchfile)
    execute('/usr/local/bin/patcha -f %s %s' % (PATCHA_IN, patchfile))

copy(target+'/tmp/keyboard', target+'/tmp/mnt/etc/default/')

######### enable root login on console if password is set
rootPassword = productProperties.get('root_password','')
if rootPassword:
    execute("sed -i '/root-login/s/#//g' %s/tmp/installation_profile" % target)

######## how to install packages

packageDependency = productProperties.get('package_dependency', 'recommends')
if packageDependency != 'recommends':
    execute("sed -i '/install-recommends/s/#//g' %s/tmp/installation_profile" % target)

######## use non-free-firmware

if nonFreeFirmware:
    execute("sed -i '/non-free-firmware/s/#//g' %s/tmp/installation_profile" % target)

if install_unattended:
    logger.notice(u"Copy profile file to initrd...")
    copy(target+'/tmp/installation_profile', target+'/tmp/mnt/preseed.cfg')
    logger.notice(u"Here comes the installation profile:")
    execute('cat %s/tmp/mnt/preseed.cfg' % (target))

scriptMessageSubject.setMessage("Rebuild patched initrd.... ")
logger.notice(u"Rebuild patched initrd.... ")
execute("rm %s/tmp/initrd" % (target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass
execute("cd %s/tmp/mnt ; find . | cpio -o -H newc > %s/tmp/initrd" % (target, target))
if nonFreeFirmware:
    if not depot.exists('%s/custom/firmware.cpio.gz' % source):
        raise Exception('firmware.cpio.gz not found in custom directory')
    else:
        depot.copy(source+'/custom/firmware.cpio.gz', target+'/tmp/')
        execute('cd '+target+'/tmp/ && gunzip firmware.cpio.gz')
        execute('mv '+target+'/tmp/initrd '+target+'/tmp/initrd.new')
        execute('cat '+target+'/tmp/initrd.new '+target+'/tmp/firmware.cpio > '+target+'/tmp/initrd')
        if os.path.isfile(target+'/tmp/initrd'):
            logger.notice(u"initrd with firmware exists")
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass

if installation_method == 'reboot':
    execute("cp %s/tmp/initrd* %s" % (target, grubTarget))
    execute("cp %s/tmp/mnt/preseed.cfg %s" % (target, grubTarget))
else:
    scriptMessageSubject.setMessage("Prepare loading kernel.... ")
    logger.notice(u"Prepare loading kernel ....")
    kernel_file = target+'/tmp/linux'
    locale += ".UTF-8"
    append_line = 'vga=788 locale=%s keymap=%s console-keymaps-at/keymap=%s auto=true priority=critical DEBCONF_DEBUG=5'% (locale,keymap,keymap)
    inird_file = target+'/tmp/initrd'
    execute("%s --load %s  --append='%s' --initrd=%s" % (which('kexec'), kernel_file, append_line, inird_file))

logger.notice(u"handle_setupAfterInstall ....")
handle_setupAfterInstall()

# Set product installation status
productOnClient.setInstallationStatus('unknown')
#productOnClient.setActionProgress('installing')
productOnClient.setActionResult('')
productOnClient.setLastAction('setup')

#scriptMessageSubject.setMessage("Writing log to service.... ")
try:
    # Write log to service
    logger.notice(u"Writing log to service")

    log = codecs.open(LOG_FILE, 'r', 'utf-8', 'replace')
    data = log.read()
    log.close()

    logger.setFileLevel(LOG_ERROR)
    if backend:
        try:
            try:
                ui.getMessageBox().addText(_(u"Writing log to service\n"))
            except:
                pass
            backend.log_write(logType = 'bootimage', data = data, objectId = clientId, append = False)
        except Exception as e:
            logger.error(u"Failed to write log to service: %s" % e)

except Exception as e:
    logger.logException(e)

if installation_method == 'reboot':
    scriptMessageSubject.setMessage("Rebooting machine ")
    logger.notice(u"Rebooting machine")
    reboot()
else:
    scriptMessageSubject.setMessage("Execute new kernel.... ")
    logger.notice(u"Execute new kernel ....")
    execute("%s -e" % (which('kexec')))
